/*
#       lidsadm.c --- The Linux Intrusion Detection System Configuration Tool 
#       (C) Huagang Xie 1999-2001 All rights reserved.
#       Portions Copyright (C) Doug Porter 2004
#       EMail complaints to xie@lids.org 
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
*/
/* $Id: lidsconf.c,v 1.3.2.19 2004/06/13 03:03:42 xie Exp $ */
/* ------------------------------------------------------------------------- */

/* Includes */

#ifdef HAVE_CONFIG_H
#include <../config.h>
#endif

/*#define _GNU_SOURCE*/
#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdarg.h>
#include <getopt.h>
/* in order to remode the warning message */
#undef __attribute_used__
#undef __attribute_pure__
#include <linux/capability.h>
#include <linux/kdev_t.h>

#include "lidstools.h"
#include "lids_capflag.h"

#include "lidsif.h"

#define LIDS_STR2(x) #x
#define LIDS_STR(X) LIDS_STR2(X)

#ifdef DEBUG
#define LIDS_DBG(msg...)  printf("LIDS." __FUNCTION__ ".l" LIDS_STR(__LINE__) ": " ##msg)
#else
#define LIDS_DBG(msg...)
#endif

//#define PRINT_ACL(acl)        printf("inode= %d, dev= %d, file= %s, type= %d, inherit= %d, o_inode= %ld, o_dev= %d, o_file= %s, from:%ld, to_%ld, state: %d\n", acl.s_ino, acl.s_dev, acl.s_file, acl.type, acl.inherit, acl.o_ino, acl.o_dev, acl.o_file, acl.time[0][0], acl.time[1][0],acl.state)

//#define PRINT_ACL(acl)        printf("subject(%d,%d,%s), type= %d, object(%ld,%d,%s), state=%d\n", acl.s_ino, acl.s_dev, acl.s_file, acl.type, acl.inherit, acl.o_ino, acl.o_dev, acl.o_file, acl.time[0][0], acl.time[1][0],acl.state)
#define PRINT_ACL(acl)	\
 	printf("%23s  %8s%s:%3d %s %20s \n\n",	\
	acl.s_file,						\
	lids_type_desc[get_type_str(abs(acl.type))],	\
	"",\
 	acl.inherit,					\
	acl.o_file,										\
 	strncmp("CAP_NET_BIND_SERVICE",acl.o_file,20)?"":disp_multi_data(acl.port,1) );

#define LIDS_WARNING(acl, string) \
{	\
	if(verbose) {\
 	fprintf(stderr, "Warning: %s\n", string); \
 	PRINT_ACL(acl); \
	}\
}
#define STRICT_ERROR(status, msg)               \
{                                               \
     if (strict) {                               \
         fprintf(stderr, "\n");                  \
         PRINT_ACL(lids[i]);                     \
         exit_error(status, msg);                \
     }                                           \
     else {                                      \
         fprintf(stderr, "Warning: %s\n", msg);  \
         PRINT_ACL(lids[i]);                     \
      }                                           \
}

/* ------------------------------------------------------------------------- */
char lids_cap_file[4][30] =
    { LIDS_GLOBAL_CAP_FILE, LIDS_BOOT_CAP_FILE, LIDS_POSTBOOT_CAP_FILE,
LIDS_SHUTDOWN_CAP_FILE };
char lids_acl_file[4][29] =
    { LIDS_GLOBAL_CONF_FILE, LIDS_BOOT_CONF_FILE, LIDS_POSTBOOT_CONF_FILE,
LIDS_SHUTDOWN_CONF_FILE };
char lids_binary_acl_file[3][29] =
    { LIDS_BOOT_ACL_FILE, LIDS_POSTBOOT_ACL_FILE, LIDS_SHUTDOWN_ACL_FILE };

/*  globals */
#define LIDS_TARGET_NUM 	8
#define LIDS_TYPE_NUM		7
#define BIG_BUFFER          1024    /* we need to dynamically calculate sizes when possible */

#define UNPROTECTED_ERROR   "You must protect the %s file %s or its directory as READONLY or DENY."
#define FULL_PATH_ERROR     "The %s file %s needs a full path."
#define GRANTED_ERROR       "The capability %s is already enabled. \n\t Disable the capablity in the cap file first."
#define DEFAULT_ERROR       "The default permission on %s or its directory already allows this."
#define RULE_EXISTS_ERROR   "Rule already in %s acl. May be linked files are using this rule."

char * get_pathname(lids_t acl, char * filename, long int inode);
char * strdup_printf (const char *format, ...);
char * get_pathname_via(char * filename, char * via_command);
void verbose_printf(const char *format, ...);

#define LIDS_INIT_FILE "/etc/lids/lids.ini"
#define LIDSADM_PATH "/sbin/lidsadm"

char lids_type_desc[LIDS_TARGET_NUM][9] =
    { "DENY", "READONLY", "APPEND", "WRITE", "IGNORE", "DISABLE", "GRANT",
"ENABLE" };

int lids_type_val[LIDS_TYPE_NUM] = { LIDS_DENY,
	LIDS_DENY | LIDS_READONLY,
	LIDS_DENY | LIDS_READONLY | LIDS_APPEND,
	LIDS_DENY | LIDS_READONLY | LIDS_APPEND | LIDS_WRITE,
	LIDS_DENY | LIDS_READONLY | LIDS_APPEND | LIDS_WRITE,
	LIDS_SOCKET,
	LIDS_CAP
};

char lids_special_target_val[2] = { 0, 1 };

struct lids_inode_acl lids_acl[1024];
lids_t eft_lids[BIG_BUFFER];
lids_t lids[BIG_BUFFER];
static int last, eft_last, last_acl;
static int lids_acl_discovery = 0;
int bindport[LIDS_PORT_ITEM][2];
int mark;
char acl_conf_file[PATH_MAX+1];
int state;
int strict = 1;
int lids_caps;
int verbose = 0;
/* for guessing */
unsigned long lids_bittab[32] = {
	0x00000001, 0x00000002, 0x00000004, 0x00000008,
	0x00000010, 0x00000020, 0x00000040, 0x00000080,
	0x00000100, 0x00000200, 0x00000400, 0x00000800,
	0x00001000, 0x00002000, 0x00004000, 0x00008000,
	0x00010000, 0x00020000, 0x00040000, 0x00080000,
	0x00100000, 0x00200000, 0x00400000, 0x00800000,
	0x01000000, 0x02000000, 0x04000000, 0x08000000,
	0x10000000, 0x20000000, 0x40000000, 0x80000000
};
static int fastguess[2048];
static int lids_search_value;
/* ------------------------------------------------------------------------- */

void
exit_error(int status, char *msg)
{
	fprintf(stderr, "lidsconf: %s\n", msg);
	if (status == 3)
		perror("reason:");
	printf("\n");
	exit(status);
}

void
exit_version()
{
	printf("lidsconf version " VERSION " for the LIDS project\n");
	exit(1);
}

void
exit_normal()
{
	printf("lidsconf version " VERSION " for the LIDS project\n"
	       "Use 'lidsconf -h' for help\n");
	exit(1);
}

void
exit_help(int more)
{
	entry_t *entry;

	printf("lidsconf version " VERSION " for the LIDS project\n"
	       "       Huagang Xie <xie@lids.org>\n"
	       "       Philippe Biondi <philippe.biondi@webmotion.net>\n\n"
	       "Usage: lidsconf -A [acl_type] [-s subject] -o object [-d] [-t from-to] [-i level] -j ACTION\n"
	       "       lidsconf -C\n"
	       "       lidsconf -D [acl_type] [-s file] [-o file] \n"
	       "       lidsconf -Z [acl_type]\n"
	       "       lidsconf -U\n"
	       "       lidsconf -L [acl_type] [-e]\n"
	       "       lidsconf -P\n"
	       "       lidsconf -S [acl_type]\n"
	       "       lidsconf -v\n"
	       "       lidsconf -[h|H]\n\n"
	       "Commands:\n"
	       "       -A,--add	To add an entry\n"
	       "       -C,--check	To check all entries\n"
	       "       -D,--delete	To delete an entry\n"
	       "       -Z,--zero	To delete all entries \n"
	       "       -U,--update	To update dev/inode numbers\n"
	       "       -L,--list	To list all entries \n"
	       "       -P,--passwd	To encrypt a password with RipeMD-160\n"
	       "       -S,--script	To write a script for all entries\n"
	       "       -v,--version	To show the version\n"
	       "       -h,--help	To list this help \n"
	       "       -H,--morehelp	To list this help with CAP/SOCKET name\n\n"
	       "subject: -s,--subject subj\n"
	       "       can be any program, must be a file\n"
	       "object: -o,--object [obj]\n"
	       "       can be a file, directory or Capability, Socket Name\n"
	       ""
	       "ACTION: -j,--jump\n"
	       "       DENY     deny access\n"
	       "       READONLY read only\n"
	       "       APPEND   append only\n"
	       "       WRITE    writable\n"
	       "       GRANT    grant capability to subject\n"
	       "       IGNORE   ignore any permissions set on this object\n"
	       "       DISABLE  disable some extersion feature\n"
	       "OPTION:\n"
	       "      -i,--inheritance Inheritance level\n"
	       "      -t,--time	Time dependency\n"
	       "      -e,--extended	Extended list\n");
	if (more) {
		printf("\nAvailable capabilities:\n");
		for_each_entry(cap_list, entry)
		    printf("%20s %s\n", entry->name, entry->desc);

		printf("\nAvailable Socket:\n");
		for_each_entry(socket_list, entry)
		    printf("%20s %s\n", entry->name, entry->desc);
	}

	exit(1);
}

void
lids_del_all()
{
	FILE *fd;

	if ((fd = fopen(acl_conf_file, "w")) == NULL) {
		exit_error(3, "cannot open acl file");
	}
	fwrite(LIDS_CONF_COMMENT, 1, strlen(LIDS_CONF_COMMENT), fd);
	fclose(fd);
}

/*
 *	lids_get_str get the ino,dev,filename from the buff.
 *	the format of buffer is "inode:dev:filename"
 */
int
lids_get_str(char *buffer, unsigned long *ino, int *dev, char *filename)
{
	char *p, *q;
	char ino_str[BIG_BUFFER];
	char ino_dev[BIG_BUFFER];

	memset(ino_str, '\0', BIG_BUFFER);
	memset(ino_dev, '\0', BIG_BUFFER);

	p = q = NULL;
	p = strchr(buffer, ':');
	if (p == NULL)
		return -1;
	*p = '\0';
	if((p - buffer) > 1024 ) {
		exit_error(2,"lids_get_str buffer overflow");
	} 
	strncpy(ino_str, buffer, p - buffer);
	p++;
	q = strchr(p + 1, ':');
	if (q == NULL)
		return -1;
	if(q-p > 1024) {
		exit_error(2,"lids_get_str buffer overflow");
	}
	strncpy(ino_dev, p, q - p);

	strncpy(filename, q + 1, PATH_MAX);
	*ino = atol(ino_str);
	*dev = atoi(ino_dev);
	return 0;
}

/*
 * check_format check if the given buf is the valid string for current version
 */
int
check_format(char *buf)
{
	int i, j;
	char *p = NULL;

	i = 0;
	p = strchr(buf, ':');
	while (p != NULL) {
		p = strchr(p + 1, ':');
		i++;
	}
	if (i != 8)
		return -1;
	return 0;
}

/*
 *	input: lids[] , include global acl and one state acl.
 *
 *  	output:	lids_effective[], include the real effective rules.
 *		
 *	side effect, warning, return the wrong rules 	
 */
void
compute_effective_rules(void)
{
	int curr_state = LIDS_STATE_GLOBAL;
	int i, default_rule, err;
    char *msg;

	eft_last = 0;

	/* should check for all the ino = 0 rule first */
	for (i = 0; i < last; i++) {
		if (lids[i].s_ino == 0) {
			err = check_same_acl(lids[i]);
			if (err < 0) {
				err = -err - 1;
 				LIDS_WARNING(lids[i], strdup_printf(RULE_EXISTS_ERROR, "this"));
				memcpy(&eft_lids[err], &lids[i],
					   sizeof (lids_t));
			}else{
			/* if difference rules found, overwrite it */
				if (err > 0) {
					err = err - 1;
 				LIDS_WARNING(lids[i], strdup_printf(RULE_EXISTS_ERROR, "Global"));
					memcpy(&eft_lids[err], &lids[i],
					       sizeof (lids_t));
				} else {
					memcpy(&eft_lids[eft_last], &lids[i],
					       sizeof (lids_t));
					eft_last++;
				}
			}
		}
	}
	while (1) {
		/* step 1.2 read the src inode != NULL's ACL */
		for (i = 0; i < last; i++) {
			if (lids[i].state == curr_state && lids[i].s_ino != 0) {
				if (*(lids[i].s_file) != '/')
                    STRICT_ERROR(2, strdup_printf(FULL_PATH_ERROR,"subject", lids[i].s_file));
				if ((default_rule = get_object_type(lids[i].s_file)) == -1)
                    STRICT_ERROR(2, strdup_printf(UNPROTECTED_ERROR, "subject", lids[i].s_file));
				if (default_rule > LIDS_READONLY)
					exit_error(2,
						   "the subject files must be protected by READONLY or DENY");
				/* check the object */
				if (lids[i].type == LIDS_CAP) {
					if (cap_raised
					    (lids_caps, lids[i].o_dev)) {
						LIDS_WARNING(lids[i], strdup_printf(GRANTED_ERROR, lids[i].o_file));
					}
				}
				if (lids[i].type != LIDS_CAP
				    && lids[i].type != LIDS_SOCKET) {
					if ((default_rule = get_object_type(lids[i].o_file)) == -1)
                        STRICT_ERROR(2, strdup_printf(UNPROTECTED_ERROR, "object", lids[i].o_file));
					if (lids[i].type < default_rule) {
						LIDS_WARNING(lids[i], strdup_printf(DEFAULT_ERROR, lids[i].o_file));
					}

					err = check_same_acl(lids[i]);
					if (err != 0) {
						err = err - 1;
						err > 0 ? err : -err;
                        LIDS_WARNING(lids[i], strdup_printf(RULE_EXISTS_ERROR, "this"));
						memcpy(&eft_lids[err], &lids[i],
						       sizeof (lids_t));
					} else {
						memcpy(&eft_lids[eft_last],
						       &lids[i],
						       sizeof (lids_t));
						eft_last++;
					}
				} else {
					/* CAP, SOCKET */
					memcpy(&eft_lids[eft_last], &lids[i],
					       sizeof (lids_t));
					eft_last++;
				}
			}
		}
		if (curr_state == state)
			break;
		else
			curr_state = state;
	}
}

/*
 *  lids read conf read /etc/lids.conf into a data structure.
 */
void
lids_do_read_conf(int curr_state)
{
	FILE *fd;
	int i = last;
	int type;
	char buffer[BIG_BUFFER];
	char *p, *q;
	char *filename;

	filename = lids_acl_file[curr_state];

	if ((fd = fopen(filename, "r")) == NULL) {
		fprintf(stderr, "acl file: %s\n",filename);
		exit_error(3, "cannot open acl file");
	}

    /*
       There are 3 variants of a conf file line.

       1) Format: "%d:%d:%s:%d:%d:%d:%d:%s:%s"
          Fields:
               s_ino: subject inode
               s_dev: subject device or filesystem
               s_file: subject file
               type: state or permission type
               inherit: inheritance level
               o_ino: object inode
               o_dev: object device or filesystem
               o_file: object file
               time: start "-" end [ "," start "-" end ... ]

       2) If object, i.e. o_file, is CAP_NET_BIND_SERVICE
          Format: "%d:%d:%s:%d:%d:%s:%d:%s:%s"
          Fields:
               s_ino: subject inode
               s_dev: subject device or filesystem
               s_file: subject file
               type: state or permission type
               inherit: inheritance level
               port: start "-" end [ "," start "-" end ... ]
               o_dev: object device or filesystem
               o_file: object file
               time: start "-" end [ "," start "-" end ... ]

       3) If object, i.e. o_file, is LIDS_SOCKET_NF_MARK
          Format: "%d:%d:%s:%d:%d:%d:%d:%s:%s"
          Fields:
               s_ino: subject inode
               s_dev: subject device or filesystem
               s_file: subject file
               type: state or permission type
               inherit: inheritance level
               mark: netfilter mark
               o_dev: object device or filesystem
               o_file: object file
               time: start "-" end [ "," start "-" end ... ]
    */
    
	while (fgets(buffer, BIG_BUFFER, fd) && i < BIG_BUFFER) {
		if (buffer[0] == '#')
			continue;
		if (strlen(buffer) >= 1024 ) {
			exit_error(2,"buffer overflow when read conf file, check the acl conf file carefully\n");
		}
		if (buffer[strlen(buffer) - 1] == '\n')
			buffer[strlen(buffer) - 1] = '\0';
		if (check_format(buffer) < 0) {
			printf("syntax error in %s\n", filename);
			exit(-1);
		}
		p = strchr(buffer, ':');
		if (p == NULL)
			continue;
		q = strchr(p + 1, ':');
		if (q == NULL)
			continue;
		p = strchr(q + 1, ':');
		if (p == NULL || *(p + 1) == ':')
			continue;
		*p = '\0';
		/* get the permission type */
		q = strchr(p + 1, ':');
		if (q == NULL)
			continue;

		*q = '\0';
		lids[i].type = atoi(p + 1);

		/* get the inherit level */
		p = strchr(q + 1, ':');
		if (p == NULL || *(p + 1) == ':')
			continue;
		*p = '\0';
		lids[i].inherit = atoi(q + 1);

		/* get the time like :19283-89283 */
		q = rindex(p + 1, ':');
		*q = '\0';

		if (lids_get_time(q + 1, lids[i].time) < 0)
			continue;

		if (lids_get_str
		    (buffer, &lids[i].s_ino, &lids[i].s_dev, lids[i].s_file) < 0)
			continue;

		if (lids_get_str
            (p + 1, &lids[i].o_ino, &lids[i].o_dev, lids[i].o_file) < 0)
			continue;

		if ((lids[i].type == LIDS_CAP)
		    && (lids[i].o_dev ==
			getentrybyname(cap_list, "CAP_NET_BIND_SERVICE")->val))
			str2data(p + 1, lids[i].port, 1, LIDS_PORT_ITEM);
		if ((lids[i].type == LIDS_SOCKET)
		    && (lids[i].o_dev == LIDS_SOCKET_NF_MARK)) {

			lids[i].mark = atoi(p + 1);
		}

        // get pathnames for conf files created by acl discovery
        strncpy(lids[i].s_file, get_pathname(lids[i], lids[i].s_file, lids[i].s_ino), PATH_MAX);
        strncpy(lids[i].o_file, get_pathname(lids[i], lids[i].o_file, lids[i].o_ino), PATH_MAX);
        
		lids[i].state = curr_state;
		i++;

	}
	last = i;

	fclose(fd);
}

/*
 *	read the capablity file from global cap and state cap
 *	merge them into a 32bits
 *
 */
void
lids_do_read_cap()
{
	int curr_state = LIDS_STATE_GLOBAL;
	int i, cap;
	char *p;
	FILE *fd;
	char buffer[BIG_BUFFER];

	lids_caps = 0;

	while (1) {
		/* read the global caps */
		if ((fd = fopen(lids_cap_file[curr_state], "r")) == NULL) {
			exit_error(3, "cannot open cap file");
		}
//              printf("ok...file=%s\n",lids_cap_file[curr_state]);
		while (fgets(buffer, BIG_BUFFER, fd)) {
			if (buffer[0] == '#')
				continue;
			if (buffer[0] != '+' && buffer[0] != '-') {
				continue;
			}
			if(strlen(buffer) >=1024 ) {
				exit_error(2, "buffer overflow when read cap file");
			}
			if (buffer[strlen(buffer) - 1] == '\n')
				buffer[strlen(buffer) - 1] = '\0';

			p = strchr(buffer, ':');
			if (p == NULL)
				continue;
			*p = '\0';
			cap = atoi(buffer + 1);
			if (cap < 0 || cap > 32) {
				fprintf(stderr,
					"WARNING: invalid cap number in cap file \n");
				continue;
			}
			if (buffer[0] == '-') {
				cap_lower(lids_caps, cap);
			} else {
				cap_raise(lids_caps, cap);
			}
		}
		if (curr_state == state)
			break;

		if (curr_state == LIDS_STATE_GLOBAL)
			curr_state = state;
	}

    verbose_printf("\t effective capability = 0x%.8x\n\n", lids_caps);

}

/*	read the cap and acl file
 *	compute the effective rules and capability
 *	check each rules for intergrity.
 */
void
lids_read_conf()
{
	eft_last = last = 0;

	if(state > LIDS_STATE_SHUTDOWN) {
		exit_error(2,"state overflow");
	}
	lids_do_read_cap();
	lids_do_read_conf(LIDS_STATE_GLOBAL);
	/* if it is not global, we will need to read it */
	if (state != LIDS_STATE_GLOBAL) {
		lids_do_read_conf(state);
	}
    compute_effective_rules();
}
void
lids_list_acl()
{
}
/* input  : lids_acl
 * output : the files 
 */
void
lids_write_acls()
{
	FILE *fd;
	int i=0,num;
	struct lids_subject_acl *s_acl;
	struct lids_object_acl *o_acl;
	struct lids_acl_header hdr;
	struct stat s_st;

	printf("Compiling %s\n",lids_binary_acl_file[state-1]);

	if(stat(LIDSADM_PATH, &s_st)) {
		exit_error(2, strdup_printf(" cannot find the lidsadm file: %s", LIDSADM_PATH));
	}

	fd = fopen(lids_binary_acl_file[state-1],"w");
	if(fd == NULL) {
		exit_error(3," can not open lids state acls\n");
	}
	
	hdr.magic = LIDS_MAGIC;
	hdr.version = LIDS_ACL_VERSION;
	hdr.sys_cap = lids_caps;
	hdr.ext_cap = 0xffffffff;
	hdr.discovery = lids_acl_discovery;
	hdr.search = lids_search_value;

	hdr.lidsadm.ino = s_st.st_ino;
	hdr.lidsadm.dev.major = major(s_st.st_dev);
	hdr.lidsadm.dev.minor = minor(s_st.st_dev);

	fwrite(&hdr,sizeof(struct lids_acl_header), 1, fd);

	while(lids_acl[i].inode.ino != 0 && i < 1024) {
		fwrite(&lids_acl[i], 1, sizeof(struct lids_inode_acl)-sizeof(lids_acl[i].name)+64,fd);
		s_acl = lids_acl[i].s_acl;
		if(s_acl == NULL ) { 
			i++;  
			continue;
		}
		num = 0;
		o_acl = s_acl->o_acl;
		while(o_acl) {
			num++;
			o_acl = o_acl->next;
		}
		/* reset it */
		o_acl = s_acl->o_acl;
		(int)s_acl->o_acl=num;
		fwrite(s_acl, 1, sizeof(struct lids_subject_acl),fd);
		while(o_acl) {
			fwrite(o_acl, 1, sizeof(struct lids_object_acl)-sizeof(o_acl->name)+64,fd);
			o_acl=o_acl->next;
		}
		i++;
	}
	fclose(fd);

	printf("Total %d ACLs\n\n",i);
	printf("================================================================================\n");
		
}
int 
get_acl_entry(unsigned int ino, unsigned int major_dev,unsigned int minor_dev)
{
	int i;
	for(i=0;i<last_acl;i++) {
		if(lids_acl[i].inode.ino == ino && 
			lids_acl[i].inode.dev.major == major_dev &&
			lids_acl[i].inode.dev.minor == minor_dev) {
			return i;
		}
	}	
	return -1;
}

int dir_relation(char *oname, char *pname)
{
	if(strlen(oname) > strlen(pname) && !strncmp(oname,pname,sizeof(pname))) {
		if(oname[strlen(pname)] == '/' ) {
			return 1;
		}
	}	
	if(strlen(oname) < strlen(pname) && !strncmp(oname,pname,sizeof(oname))) {
		if(pname[strlen(oname)] == '/' ) {
			return 2;
		}
	}	
	return 0;
}
/*
*/
int
get_last_perm(struct lids_inode_acl *i_acl)
{
	int i =0;
	while(i_acl->perm[i].sid != 0) {
		i++;
	}
	return i;
}
/* 

*/
void 
merge_perm(struct lids_inode_acl *d_i_acl, struct lids_inode_acl *s_i_acl)\
{
	int i,j=0;
	int last_d_acl, last_s_acl;
	

	last_s_acl = get_last_perm(s_i_acl);

	for(i=0;i<last_s_acl;i++) {
		j=0;
		while(d_i_acl->perm[j].sid !=0 && j<64){
			if(d_i_acl->perm[j].sid == s_i_acl->perm[i].sid && 
				d_i_acl->perm[j].oid == s_i_acl->perm[i].oid )
				break;
			j++;
		}
		if(j==64) {
			exit_error(2,"Fatal: too many perm for the file\n");
		}
		if(d_i_acl->perm[j].type < s_i_acl->perm[i].type) {
			memcpy(&d_i_acl->perm[j], &s_i_acl->perm[i], sizeof(struct lids_perm));
		}
	}
	last_d_acl = get_last_perm(d_i_acl);
	
}

/*
	self container
*/
void
lids_add_obj(int sid, struct lids_object_acl *o_acl)
{
	int ent;
	int i;

	ent = get_acl_entry(o_acl->inode.ino, o_acl->inode.dev.major,o_acl->inode.dev.minor);
	if(ent<0) {
		ent = last_acl; last_acl ++; 
		memset(&lids_acl[ent],0,sizeof(struct lids_inode_acl));
		lids_acl[ent].inode.ino = o_acl->inode.ino;
		lids_acl[ent].inode.dev.major = o_acl->inode.dev.major;
		lids_acl[ent].inode.dev.minor = o_acl->inode.dev.minor;
		/* READONLY type first */
		lids_acl[ent].type  = 1 ; /* FIXME, get it from name */
		lids_acl[ent].s_acl == NULL;
		lids_acl[ent].magic = LIDS_MAGIC;
		strncpy(lids_acl[ent].name,o_acl->name, strlen(o_acl->name));
		memset(lids_acl[ent].perm, 0, 64*sizeof(struct lids_perm));
	}
	/* add the lids_perm */
	i = 0;
	while(lids_acl[ent].perm[i].sid!=0 && i<64) {
		if((lids_acl[ent].perm[i].sid == sid )&& lids_acl[ent].perm[i].oid == o_acl->oid) break;
		i++;
	}
	if(i==64) {
		exit_error(2,"Fatal, Object's perm exceed max number 64\n");
	}

	/* by default type is 0, and if exist, compare the type */
	if(lids_acl[ent].perm[i].type < o_acl->type) {
		lids_acl[ent].perm[i].sid = sid;
		lids_acl[ent].perm[i].oid = o_acl->oid;
		lids_acl[ent].perm[i].type= o_acl->type;
	}
	
	/* check all the lids_acl entry */
	for(i=0;i<last_acl;i++) {
		switch (dir_relation(lids_acl[ent].name, lids_acl[i].name)) {
	/* look for its parent dir, if found, adding its parent's perm 
		to it*/
		/* if we are ts subdir ,add current to the sudri */
		case 1:
			merge_perm(&lids_acl[ent],&lids_acl[i]);
			break;
		case 2:
			merge_perm(&lids_acl[i],&lids_acl[ent]);
			break;
		default:
			break;
		}
	/* look for its sub-dir if it is directory , if found, add 
		current perm to it */
	}
}
/* LIDS guessiong sort routine */

static int 
mkdev(int major, int minor)
{
	return (major<<8)|(minor);
}

static int 
lids_inode_dev_ino_cmp(struct lids_inode_acl *s1,
					     struct lids_inode_acl *s2)
{
	unsigned long s1_dev = mkdev(s1->inode.dev.major, s1->inode.dev.minor);
	unsigned long s2_dev = mkdev(s2->inode.dev.major, s2->inode.dev.minor);
	unsigned long s1_ino = s1->inode.ino;
	unsigned long s2_ino = s2->inode.ino;

	if (s1_dev < s2_dev)
		return -1;
	if (s1_dev > s2_dev)
		return 1;
	if (s1_ino < s2_ino)
		return -1;
	if (s1_ino > s2_ino)
		return 1;
	return 0;		/* Aieeeh, duplicate entry */
}
static int 
lids_sort_acls()
{
	int i;
	int working;
	struct lids_inode_acl tmp;

	/* sort inode entries by device and inode */

	do {
		working = 0;
		for (i = 0; i < last_acl- 1; i++) {
			switch (lids_inode_dev_ino_cmp(&(lids_acl[i]),
						       &(lids_acl[i + 1]))) {
			case 0:
				return 0;
			case 1:
				tmp = lids_acl[i];
				lids_acl[i] = lids_acl[i + 1];
				lids_acl[i + 1] = tmp;
				working = 1;
			}
		}
		for (i = last_acl - 1; i > 0; i--) {
			switch (lids_inode_dev_ino_cmp(&(lids_acl[i - 1]),
						       &(lids_acl[i]))) {
			case 0:
				return 0;
			case 1:
				tmp = lids_acl[i - 1];
				lids_acl[i - 1] = lids_acl[i];
				lids_acl[i] = tmp;
				working = 1;
			}
		}
	} while (working);

	/* set up base index for fast search: no, we can't use >0x20000000
	   here but anyway it'll take some years for computers to
	   allow for the memory required for this size :-) */

	lids_search_value = 0x20000000;
	while (last_acl <= lids_search_value && lids_search_value)
		lids_search_value >>= 1;

	return 1;
}

void
lids_init_fastguess(int inode, int major, int minor)
{

	int i;
	long dev;

	dev = mkdev(major,minor);
	i = (dev ^ inode) & 0xffff;
	fastguess[i >> 5] |= lids_bittab[i & 31];
}

/*
	input the lids[]
	output:
		add/modify new entry in lids[] which have
		have lids_perm  strucutre.
*/
void
lids_gen_perm()
{
	int i;
	struct lids_object_acl *o_acl;

	for(i=1;i<last_acl;i++) {
		if(lids_acl[i].s_acl && lids_acl[i].s_acl->o_acl) {
			o_acl = lids_acl[i].s_acl->o_acl; 
			while(o_acl) {
				lids_add_obj(lids_acl[i].s_acl->sid, o_acl);	
				o_acl = o_acl->next;
			}
		} 
	}	
}
void
lids_read_ini_file()
{
	FILE *fd;
	char buffer[2048];
	int i=0;

	fd = fopen(LIDS_INIT_FILE, "r");
	if(!fd) {
		fprintf(stderr,"open %s failed\n",LIDS_INIT_FILE);
		return;
	}
	while (fgets(buffer, 1024, fd) && i < 1024) {
		if (buffer[0] == '#')
			continue;
		if (buffer[strlen(buffer) - 1] == '\n')
			buffer[strlen(buffer) - 1] = '\0';
		if(!strncmp(buffer,"ACL_DISCOVERY", strlen("ACL_DISCOVERY"))) {
			if(strchr(buffer,'1')==NULL) {	
				printf("LIDS: ACL Discovery is OFF\n");
				continue;
			}
			lids_acl_discovery = 1;	
			printf("LIDS: ACL Discovery is ON\n");
		}
		i++;
	}
	fclose(fd);
}
/* input , the elf_lids()
   ouput,  the 3 file in binary format which will used by the kernel 
*/
void
lids_compile()
{
	int i;
	int err = 0;
	struct lids_subject_acl *s_acl;
	struct lids_object_acl *o_acl;
	unsigned int ino, dev;
	int ent;
	char *pname;

	last_acl = 0;
	memset(lids_acl, 0, sizeof(lids_acl));
	/* for all the eft_lids, merge them into one big error */
	for (i = 0; i < eft_last; i++) {
		if(eft_lids[i].s_ino == 0 ) {
			ino = eft_lids[i].o_ino;
			dev = eft_lids[i].o_dev; 
			pname = eft_lids[i].o_file;
		} else {
			ino = eft_lids[i].s_ino;
			dev = eft_lids[i].s_dev; 	
			pname = eft_lids[i].s_file;
		}
		ent = get_acl_entry(ino,major(dev),minor(dev));
		if(ent<0) {
			ent = last_acl; last_acl ++; 
			memset(&lids_acl[ent],0,sizeof(struct lids_inode_acl));
			lids_acl[ent].inode.ino = ino;
			lids_acl[ent].inode.dev.major = major(dev);
			lids_acl[ent].inode.dev.minor = minor(dev);
			/* READONLY type first */
			lids_acl[ent].type  = 1 ;
			lids_acl[ent].s_acl == NULL;
			lids_acl[ent].magic = LIDS_MAGIC;
			strncpy(lids_acl[ent].name,pname,strlen(pname));
			/* guessing array */
			lids_init_fastguess(ino, major(dev), minor(dev));
		}
		if(eft_lids[i].s_ino == 0 ) {
			lids_acl[ent].type = eft_lids[i].type;
		} else {
			s_acl = lids_acl[ent].s_acl; 
			if(!s_acl) {
				s_acl = malloc(sizeof(struct lids_subject_acl));
				if(!s_acl) 
					exit_error(2,"malloc for lids_subject_acl failed\n");
				memset(s_acl,0, sizeof(struct lids_subject_acl));
				lids_acl[ent].s_acl = s_acl;
				lids_acl[ent].s_acl->sid = ent+1;	/* the sid > 0*/
			}

			switch(eft_lids[i].type) {
			case LIDS_CAP:	
				cap_raise(s_acl->sys_cap, eft_lids[i].o_dev);
				s_acl->cap_inherit[eft_lids[i].o_dev].inherit = eft_lids[i].inherit;
				/* the port */
				memcpy(s_acl->port, eft_lids[i].port, LIDS_PORT_ITEM * 2 * sizeof (int *));
				break;
			case LIDS_SOCKET: 
				cap_lower(s_acl->ext_cap, eft_lids[i].o_dev);
				break;
			default:
				o_acl = malloc(sizeof(struct lids_object_acl));
				if(!o_acl) {
					exit_error(2,"malloc for lids_object_acl failed\n");
				}
				o_acl->type = eft_lids[i].type;
				o_acl->inherit = eft_lids[i].inherit;
				o_acl->inode.ino = eft_lids[i].o_ino;
				o_acl->inode.dev.major = major(eft_lids[i].o_dev);
				o_acl->inode.dev.minor = minor(eft_lids[i].o_dev);
				strncpy(o_acl->name, eft_lids[i].o_file, strlen(eft_lids[i].o_file));
				
				/* sid */
				o_acl->sid = s_acl->sid;
				/* oid */
				if(s_acl->o_acl == NULL) {
					o_acl->oid = 1;
				}else{
					o_acl->oid = s_acl->o_acl->oid++;
				}
				/* FIXME, the port*/
				o_acl->next = s_acl->o_acl;
				s_acl->o_acl = o_acl; 
				break;
			}
		}
			
	}
	lids_sort_acls();
	lids_gen_perm();
	lids_read_ini_file();
	/* write to the file */
	lids_write_acls();		
}

/*
 *	performace the ACL rules interigitry checking
 *
 */
void
lids_check()
{
	for (state = LIDS_STATE_GLOBAL; state <= LIDS_STATE_SHUTDOWN; state++) {
		printf("reading %s\n", lids_acl_file[state]);
		lids_read_conf();
		if(state>LIDS_STATE_GLOBAL) {
			lids_compile();
		}
	}

}

void
lids_write_file()
{
	int i = 0;
	FILE *fd;

	if ((fd = fopen(acl_conf_file, "w")) == NULL) {
		exit_error(3, "cannot open acl file");
	}
	fwrite(LIDS_CONF_COMMENT, 1, strlen(LIDS_CONF_COMMENT), fd);

	for (i = 0; i < last; i++) {

		if (lids[i].state == state && lids[i].type != LIDS_DELETE_FLAG) {
			if (lids[i].type == LIDS_CAP &&
			    lids[i].o_dev == getentrybyname(cap_list,
							    "CAP_NET_BIND_SERVICE")->
			    val)

				fprintf(fd, "%d:%d:%s:%d:%d:%s:%d:%s:%s\n",
					lids[i].s_ino, lids[i].s_dev,
					lids[i].s_file, lids[i].type,
					lids[i].inherit,
					disp_multi_data(lids[i].port, 1),
					lids[i].o_dev,
					lids[i].o_file,
					disp_multi_data(lids[i].time, 0), 0);
			else
				fprintf(fd, "%d:%d:%s:%d:%d:%d:%d:%s:%s\n",
					lids[i].s_ino, lids[i].s_dev,
					lids[i].s_file, lids[i].type,
					lids[i].inherit,
					lids[i].o_ino, lids[i].o_dev,
					lids[i].o_file,
					disp_multi_data(lids[i].time, 0), 0);
		}
	}

	fclose(fd);

}

/*
 *	get the str desc number from type number
 */
int
get_type_str(int type)
{
	int k, j;

	k = -1;
	for (j = 0; j < LIDS_TYPE_NUM; j++) {
		if (type == lids_type_val[j]) {
			k = j;
			break;
		}
	}
	if (k == -1)
		exit_error(2, "type mismatch in acl file");
	return k;
}

void
lids_list_file(int type)
{
	int i, j, k;
	static char anyfile[] = "Any file";
	char *src, *obj;

	lids_read_conf();

	if (type == LIDS_EXTEND)
		printf("Subj ino,dev   Obj ino,dev  type  ");
	printf
	    ("                Subject   ACCESS  inherit		Object\n");

	printf
	    ("----------------------------------------------------------------------------\n");
	for (i = 0; i < last; i++) {
		if (lids[i].state == state) {
			src = anyfile;
			obj = anyfile;

			if (*lids[i].s_file)
				src = lids[i].s_file;

			if (*lids[i].o_file)
				obj = lids[i].o_file;

			if (type == LIDS_EXTEND)
				printf("%6d,%6d %6d,%6d %6d",
				       lids[i].s_ino, lids[i].s_dev,
				       lids[i].o_ino, lids[i].o_dev,
				       lids[i].type);

			printf("%23s  %8s%s:%3d  %20s %s \n",
			       src,
			       lids_type_desc[get_type_str(abs(lids[i].type))],
			       "",
			       lids[i].inherit,
			     //  time2str(lids[i].time),
			       obj,
			       strncmp("CAP_NET_BIND_SERVICE", obj,
				       20) ? "" : disp_multi_data(lids[i].port,
								  1)
			      // strncmp("LIDS_SOCKET_NF_MARK", obj,
				//       19) ? 0 : lids[i].mark
				);
		}
	}
	printf("\n\n");
}

void
lids_write_script()
{
	static char script_ext[] = ".sh";
    FILE *fd;
    int i, type, inherit;
	char *acl_script_file, *stateString, *src, *obj, *access;

    strict = 0;
	lids_read_conf();
            
    switch (state) {
        case LIDS_STATE_BOOT:
            stateString = " BOOT";
            break;
        case LIDS_STATE_POSTBOOT:
            stateString = " POSTBOOT";
            break;
        case LIDS_STATE_SHUTDOWN:
            stateString = " SHUTDOWN";
            break;
        default:
            stateString = "";
    }
    
    acl_script_file = (char *) malloc(strlen(acl_conf_file) + strlen(script_ext) + 1);
    sprintf(acl_script_file, "%s%s", acl_conf_file, script_ext);
	if ((fd = fopen(acl_script_file, "w")) == NULL) {
		exit_error(3, "cannot open acl file");
	}
    
	fwrite(LIDS_SCRIPT_HEADER, 1, strlen(LIDS_SCRIPT_HEADER), fd);

    fprintf(fd, "lidsconf -Z%s\n", stateString);
    
	for (i = 0; i < last; i++) {
		if (lids[i].state == state) {

			fprintf(fd, "lidsconf -A");
            
            fprintf(fd, "%s", stateString);
            
            src = lids[i].s_file;
			if (*src) {
                fprintf(fd, " -s %s", src);
            }

            obj = lids[i].o_file;
			if (*obj)
                fprintf(fd, " -o %s", obj);
            
            inherit = lids[i].inherit;
            if (inherit > 0) 
                fprintf(fd, " -i %d", inherit);
            
            type = abs(lids[i].type);
            if (type >= 0) {
                access = lids_type_desc[get_type_str(abs(lids[i].type))]; 
                fprintf(fd, " -j %s", access);
            }
            
			fprintf(fd, "\n");
            
		}
	}
    
    fclose(fd);
    
    printf("Wrote%s script to %s\n", stateString, acl_script_file);
    free(acl_script_file);
}

/*
 *  get pathname for an executable.
 */
char * get_pathname(lids_t acl, char * filename, long int inode)
{
    char *pathname;

    // ignore CAP rules, missing filenames, and files with existing pathnames
    if (acl.type != LIDS_CAP &&
        strlen(filename) > 0 &&
        *filename != '/') {

        // for speed, try "which" first
        // This works well for subjects and some objects, although it can return the wrong
        // pathname because the pathname was different when acl discovery was running 
        pathname = get_pathname_via(filename, strdup_printf("which %s", filename));
        
        // if that didn't work, try "find -inum <inode>
        // This works for objects as well as subjects, but is very, very slow.
        // we could get the wrong pathname if an inode matches on the wrong file system. 
        // And the pathname may look wrong if there are multiple links to the inode,
        // although it will still work.
        if (strcmp(filename, pathname) == 0 &&
            inode > 0) {
            fprintf(stderr, "Finding inode %d for %s...\n", inode, filename);
            pathname = get_pathname_via(filename, strdup_printf("find / -inum %d", inode));
            fprintf(stderr, "Pathname for inode %d is %s\n", inode, pathname);
        }
        
        // if the end of the pathname doesn't match the filename
        if (strcmp(filename, &(pathname[strlen(pathname) - strlen(filename)])) != 0) {
            fprintf(stderr, "Not using bad pathname %s for filename %s.\n", pathname, filename); 
            pathname = filename;
        }
    
    }
    else {
        pathname = filename;
    }
    
    return pathname;
}

/*
 *  get pathname for a file via the specified command.
 */
char * get_pathname_via(char * filename, char * via_command)
{
    static char command_template[] = "%s > %s";
	static char tmp_file[] = "/tmp/which.out";
    char *command, pathname[PATH_MAX + 1], *s;
    FILE *fd;

    // issue command
    command = (char *) malloc(strlen(command_template) + 
                              strlen(via_command) + 
                              strlen(tmp_file) + 1);
    sprintf(command, command_template, via_command, tmp_file);
    system(command);
    free(command);
    
    // read pathname
    if ((fd = fopen(tmp_file, "r")) == NULL) {
        exit_error(3, "cannot open tmp file");
    }
    if (fgets(pathname, PATH_MAX, fd)) {
        
        if (strlen(pathname) > 0) {
            
            // strip trailing white space
            s = &(pathname[strlen(pathname) - 1]);
            while (s > pathname && isspace(*s)) {
                *s = '\0';
                --s;
            }
    
            // nobody is going to free this, but it won't use enough space to matter                           
            filename = strdup(pathname);
        }
        
    }

    fclose(fd);
    unlink (tmp_file);
    
    return filename;
}

/*
 *	lids_search_file()
 *
 *	lids search the file by given file and type.
 *
 */
int
lids_search_file(char *s_file, char *o_file)
{
	int i, flag, number = 0;

	for (i = 0; i < last; i++) {
		flag = 0;
		if (*s_file != '\0') {
			flag = strcmp(lids[i].s_file, s_file);
		}
		if (*o_file != '\0')
			flag |= strcmp(lids[i].o_file, o_file);
		if (!flag) {
			number++;
			lids[i].type = LIDS_DELETE_FLAG;
		}
	}
	return number;
}

/*
 *	lids_search_inode()
 *
 *	lids search the file by given inode , dev  and type.
 *
 */
int
check_same_acl(lids_t acl)
{
	int i;
	int err = 0;

	for (i = 0; i < eft_last; i++) {
		if (eft_lids[i].s_ino == acl.s_ino &&
		    eft_lids[i].s_dev == acl.s_dev &&
		    eft_lids[i].o_ino == acl.o_ino &&
		    eft_lids[i].o_dev == acl.o_dev) {
			if (eft_lids[i].state == acl.state) {
				err = -(i + 1);
				break;
			} else {
				err = (i + 1);
			}
		}
	}
	return err;
}

void
lids_del_file(char *subject_file, char *object_file)
{
	int i;
	char canonical_s_path[PATH_MAX+16];
	char canonical_o_path[PATH_MAX+16];

	lids_read_conf();

	i = lids_search_file(subject_file, object_file);
	if(strlen(subject_file) > PATH_MAX) {
		exit_error(2,"subject file length too long ");
	}

	if (!i) {
		realpath(subject_file, canonical_s_path);
		realpath(object_file, canonical_o_path);
		i = lids_search_file(canonical_s_path, canonical_o_path);
		if (!i)
			exit_error(2, "the file does not exit in acl file");
	}
	printf("Deleted %d ACL\n", i);

	lids_write_file();
}

/*
 *	get special type from typestr, when lidsadm -A -s ..-o .. -t -j ..
 */

int
get_object_type(char *filename)
{
	int i;
	char o_file[PATH_MAX + 1];
	char *p = NULL;
	char root[2];
	char *fileptr;

	if(strlen(filename) > 1023) {
		exit_error(2,"filename overflow in get_object_type");
	}
	strncpy(o_file, filename,PATH_MAX);
	strncat(o_file, "/",1);
	root[0] = '/';
	root[1] = '\0';

	while (p != o_file) {
		p = rindex(o_file, '/');
		if (!p)
			break;
		*p = '\0';
		if (*o_file != '\0') {
			fileptr = o_file;
		} else {
			fileptr = root;
		}
		for (i = 0; i < eft_last; i++) {
			if (eft_lids[i].s_ino == 0
			    && strcmp(fileptr, eft_lids[i].o_file) == 0) {
				return eft_lids[i].type;
			}
		}
	}
	return -1;

}

/*
 *	lids_add_file 
 */
void
lids_add_file(char *subject_file, char *object_file, int type, int inherit,
	      time_t timescale[][2])
{
	struct stat s_st, o_st, z_st;
	char canonical_s_path[PATH_MAX+16];
	char canonical_o_path[PATH_MAX+16];
	int default_rule = -1;
	int sys_cap = 0;	/* 1 for system cap, 0 for individual cap */
	entry_t *entry;
	lids_t acl;
	int err;

	lids_read_conf();

	if (last >= BIG_BUFFER) {
		exit_error(2, strdup_printf("cannot exceed %d entries.", BIG_BUFFER));
	}

	LIDS_DBG("sub=[%s],type=%d\n", subject_file, type);

	if (*subject_file == '\0') {
		sys_cap = 1;
	}

	if (!sys_cap) {
		realpath(subject_file, canonical_s_path);
		if (stat(canonical_s_path, &s_st)) {
            exit_error(2, strdup_printf(" cannot find the subject file: %s", subject_file));
		}
		if (!S_ISREG(s_st.st_mode)) {
            exit_error(2, strdup_printf(" subject file needs to be a regular file: %s", subject_file));
		}

		/* check if the subject file is executable or not */
	} else {
		s_st.st_ino = s_st.st_dev = 0;
	}
	if (type == LIDS_CAP) {
		o_st.st_ino = -1;	/* special type's ino */
		if (entry = getentrybyname(cap_list, object_file)) {
			o_st.st_dev = entry->val;
			if (cap_raised(lids_caps, o_st.st_dev)) {
				fprintf(stderr,	
					   "Warning: The capability grant is already enable in some of the state, this acl will not be effective in these states. Disable the capablity in the cap file first.");
			}
		} else {
			exit_error(2,
				   "special type must be one of CAP_XXX_XXX. use lidsadm -h for details.");
		}
		strncpy(canonical_o_path, object_file,PATH_MAX);
	} else if (type == LIDS_SOCKET) {
		o_st.st_ino = -1;	/* special type's ino */
		if (entry = getentrybyname(socket_list, object_file)) {
			o_st.st_dev = entry->val;
		} else {
			exit_error(2,
				   "special type must be one of LIDS_SOCKET_XXX_XXX. use lidsconf -h for details.");
		}
		if (o_st.st_dev == LIDS_SOCKET_NF_MARK) {
			o_st.st_ino = mark;
		}
		strncpy(canonical_o_path, object_file,PATH_MAX);
	} else {
		realpath(object_file, canonical_o_path);
		if (stat(canonical_o_path, &o_st)) {
	              exit_error(2, strdup_printf(" cannot find the object file: %s", object_file));
		}
	}
	/* fill up the ACL */
	acl.s_ino = s_st.st_ino;
	acl.s_dev = s_st.st_dev;
	acl.o_ino = o_st.st_ino;
	acl.o_dev = o_st.st_dev;

	acl.type = type;
	acl.inherit = inherit;
	acl.state = state;

	strncpy(acl.s_file, canonical_s_path,PATH_MAX);
	strncpy(acl.o_file, canonical_o_path,PATH_MAX);

	memcpy(acl.time, timescale, LIDS_TIME_ITEM * 2 * sizeof (int *));
	memcpy(acl.port, bindport, LIDS_PORT_ITEM * 2 * sizeof (int *));
	/* check the subject */
	if (!sys_cap) {
		if ((default_rule = get_object_type(acl.s_file)) == -1)
 			exit_error(2, strdup_printf(UNPROTECTED_ERROR, "subject", acl.s_file));
		if (default_rule > LIDS_READONLY)
 			exit_error(2, strdup_printf(GRANTED_ERROR, acl.o_file));
	}
	/* check the object */
	if (type != LIDS_CAP && type != LIDS_SOCKET) {
		if (*(acl.o_file) != '/')
             		exit_error(2, strdup_printf(FULL_PATH_ERROR, "object", acl.o_file));
		if ((default_rule = get_object_type(acl.o_file)) == -1) {
			if (!sys_cap)
 				exit_error(2, strdup_printf(UNPROTECTED_ERROR, "object", acl.o_file));
		}
		/* if current type have less persmission of defualt _rule */
		if (type < default_rule && !sys_cap) {
 			exit_error(2, strdup_printf(DEFAULT_ERROR, acl.o_file));
		}
	}

	LIDS_DBG("canonical_o_file=%s,o_ino =%d\n", canonical_o_path,
		 o_st.st_ino);

	err = check_same_acl(acl);

	if (err < 0)
        // why is this a fatal error when it's usually a warning?
        exit_error(2, strdup_printf(RULE_EXISTS_ERROR, "this"));
	if (err > 0)
        LIDS_WARNING(acl, strdup_printf(RULE_EXISTS_ERROR, "Global"));

	memcpy(&lids[last], &acl, sizeof (lids_t));

	last++;
	lids_write_file();
}

void
lids_update()
{
	struct stat s_st, o_st;
	int i;
	lids_read_conf();

	for (i = 0; i < last; i++) {
		/*
		 *      check subject files
		 */
		if (*lids[i].s_file != '\0') {
			if (stat(lids[i].s_file, &s_st)) {
				printf
				    ("%s doesn't exist anymore. not removed.\n",
				     lids[i].s_file);
			} else {
				if ((s_st.st_ino != lids[i].s_ino)
				    || (s_st.st_dev != lids[i].s_dev)) {
					printf
					    ("subject file %s was (%u:%u inode %lu) instead of (%u:%u %lu). corrected.\n",
					     lids[i].s_file,
					     MAJOR(lids[i].s_dev),
					     MINOR(lids[i].s_dev),
					     lids[i].s_ino,
					     MAJOR((unsigned long) s_st.st_dev),
					     MINOR((unsigned long) s_st.st_dev),
					     s_st.st_ino);
					lids[i].s_ino = s_st.st_ino;
					lids[i].s_dev = s_st.st_dev;
				}
			}
		}
		/*
		 *      check object files
		 */
		if (*lids[i].o_file != '\0' && stat(lids[i].o_file, &o_st)) {
			if ((getentrybyname(cap_list, lids[i].o_file) < 0)
			    && lids[i].type != LIDS_CAP
			    && lids[i].type != LIDS_SOCKET) {
				printf
				    ("%s doesn't exist anymore. not removed.\n",
				     lids[i].o_file);
			}
		} else {
			if (o_st.st_ino != lids[i].o_ino
			    || o_st.st_dev != lids[i].o_dev) {
				printf
				    ("object file %s was (%u:%u inode %lu) instead of (%u:%u %lu). corrected.\n",
				     lids[i].o_file, MAJOR(lids[i].o_dev),
				     MINOR(lids[i].o_dev), lids[i].o_ino,
				     MAJOR((unsigned long) o_st.st_dev),
				     MINOR((unsigned long) o_st.st_dev),
				     o_st.st_ino);

				lids[i].o_ino = o_st.st_ino;
				lids[i].o_dev = o_st.st_dev;
			}
		}
	}
	lids_write_file();
}
void
lids_make_passwd()
{
	char passwd[1024];
	int times = 0;
	int ok;
	FILE *fd;

	memset(passwd, '\0', 1024);
	while (times < 3) {
		times++;
		if ((ok = read_crypto_passwd(passwd, 1, 2)) == 0)
			break;
	}

	if (ok == 0) {
		fd = fopen(LIDS_PW_FILE, "w");
		if (fd == NULL)
			return;
		fwrite(passwd, 1, strlen(passwd), fd);
		fclose(fd);
		printf("wrote password to %s\n", LIDS_PW_FILE);
	}

}

void
get_acl_type(int argc, char **argv)
{
	char acl_type_str[16];

	memset(acl_type_str, 0, 16);
	if (optind < argc) {
		if(strlen(argv[optind]) > 16) {
			exit_error(2,"Fatal: acl type string overly long");
		}
		strncpy(acl_type_str, argv[optind], 16);
	}
	if (strlen(acl_type_str) == 0 || acl_type_str[0] == '-') {
		strncpy(acl_conf_file, LIDS_GLOBAL_CONF_FILE,PATH_MAX);
		state = LIDS_STATE_GLOBAL;
	} else if (strncmp(acl_type_str, "POSTBOOT", 8) == 0) {
		state = LIDS_STATE_POSTBOOT;
		strncpy(acl_conf_file, LIDS_POSTBOOT_CONF_FILE,PATH_MAX);
		optind++;
	} else if (strncmp(acl_type_str, "BOOT", 4) == 0) {
		state = LIDS_STATE_BOOT;
		strncpy(acl_conf_file, LIDS_BOOT_CONF_FILE,PATH_MAX);
		optind++;
	} else if (strncmp(acl_type_str, "SHUTDOWN", 8) == 0) {
		state = LIDS_STATE_SHUTDOWN;
		strncpy(acl_conf_file, LIDS_SHUTDOWN_CONF_FILE,PATH_MAX);
		optind++;
	} else {
		printf("acl_type_str [%s] len = %d\n", acl_type_str,
		       strlen(acl_type_str));
		exit_error(2, "Must be BOOT, POSTBOOT , SHUTDOWN or NULL");
	}
 	verbose_printf("Using ACL FILE: %s\n", acl_conf_file);
}
/* Duplicate a string containing formatted output.  */
char * strdup_printf (const char *format, ...) 
{
	char *s = (char *) malloc(BIG_BUFFER + 1);
	char *t;
	va_list arg;  
	    
	s = (char *)malloc(BIG_BUFFER+1);
	if(!s) {
		exit_error(2,"malloc error\n");
	}
	memset(s,0,BIG_BUFFER+1);

	va_start (arg, format);   
	vsnprintf (s, BIG_BUFFER, format, arg);   
	va_end (arg);
	t = strdup(s);
	free (s);
    
	return t; 
}

void verbose_printf(const char *format, ...) 
{
    va_list arg;   

    if (verbose) {    
        va_start (arg, format);   
        vprintf (format, arg);  
        va_end (arg);
    }
}

static char shortopts[] = "U::A::D::Z::L::SCPs:o::j:i:dehHt:vV";

static struct option longopts[] = {
	{"update", 2, 0, 'U'},
	{"add", 2, 0, 'A'},
	{"delete", 2, 0, 'D'},
	{"zero", 2, 0, 'Z'},
	{"list", 2, 0, 'L'},
 	{"script", 0, 0, 'S'},
	{"passwd", 0, 0, 'P'},
	{"subject", 1, 0, 's'},
	{"object", 2, 0, 'o'},
	{"jump", 1, 0, 'j'},
	{"inheritance", 1, 0, 'i'},
	{"extended", 0, 0, 'e'},
	{"time", 1, 0, 't'},
	{"help", 0, 0, 'h'},
	{"morehelp", 0, 0, 'H'},
	{"version", 0, 0, 'v'},
	{"verbose", 0, 0, 'V'},
	{"check", 0, 0, 'C'},
	{0, 0, 0, 0}
};

main(int argc, char **argv)
{
	int command = LIDS_NONE;
	int type = LIDS_NONE;
	char subject_file[PATH_MAX + 1], object_file[PATH_MAX + 1];
	char sw_flag[16];
	char lids_time[176];
	char acl_type_str[16];
	int c, i;
	int index = 0;
	int cap_type = 0;	/* program capability type */
	int inherit = 0;	/* if the acl is inheritable */
	time_t timescale[LIDS_TIME_ITEM][2];

	setentry(cap_list);
	setentry(socket_list);
	memset(subject_file, '\0', PATH_MAX + 1);
	memset(object_file, '\0', PATH_MAX + 1);
	memset(timescale, 0, sizeof (timescale));
	if (getuid() != 0) {
		exit_error(2, "you must be root to run this program");
	}
	while ((c = getopt_long(argc, argv, shortopts, longopts, &index)) != -1) {
		switch (c) {
		case 'U':
			if (command != LIDS_NONE)
				exit_error(2, "multiple commands specified");
			command = LIDS_UPDATE;
			get_acl_type(argc, argv);
			break;
		case 'A':
			if (command != LIDS_NONE)
				exit_error(2, "multiple commands specified");
			command = LIDS_ADD;
			get_acl_type(argc, argv);
			break;
		case 'D':
			if (command != LIDS_NONE)
				exit_error(2, "multiple commands specified");
			command = LIDS_DELETE;
			get_acl_type(argc, argv);
			break;
		case 'L':
			if (command != LIDS_NONE)
				exit_error(2, "multiple commands specified");
			command = LIDS_LIST;
			get_acl_type(argc, argv);
			break;
		case 'S':
			if (command != LIDS_NONE)
				exit_error(2, "multiple commands specified");
			command = LIDS_SCRIPT;
			get_acl_type(argc, argv);
			break;
		case 'e':
			if (command != LIDS_LIST)
				exit_error(2, "error commands specified");
			type = LIDS_EXTEND;
			break;
		case 'Z':
			if (command != LIDS_NONE)
				exit_error(2, "multiple commands specified");
			command = LIDS_DELETE_ALL;
			get_acl_type(argc, argv);
			break;
		case 'o':	/* for object */
			if (command == LIDS_NONE)
				exit_error(2, "no comands specified");

			if (optind < argc) {
				if(strlen(argv[optind+1]) > 1024) {
					exit_error(2, "file length too long, > 1024");
				}
				strncpy(object_file, argv[optind++],PATH_MAX);
			}
			if (!strcmp(object_file, "CAP_NET_BIND_SERVICE")) {
				if (optind != argc) {

					if (str2data
					    (argv[optind], bindport, 1,
					     LIDS_PORT_ITEM) < 0)
						exit_error(2,
							   "The binding port is invalid");
					optind++;
				} else
					bindport[0][0] = bindport[0][1] = 0;
			}
			if (!strcmp(object_file, "LIDS_SOCKET_NF_MARK")) {
				exit_error(2, "LIDS_SOCKET_NF_MARK is not supported in this version");
			}
			if (strchr(object_file, ':'))
				exit_error(2, "filename can not contain :");
			break;
		case 's':	/* program capability subject */
			if (command == LIDS_NONE)
				exit_error(2, "no comands specified");
			if(strlen(optarg) > 1024) {
				exit_error(2, "file length too long, > 1024");
			}
			strncpy(subject_file, optarg,PATH_MAX);
			if (strchr(object_file, ':'))
				exit_error(2, "filename must not contain :");
			break;
		case 't':	/* time scale restrition */
			if (command == LIDS_NONE)
				exit_error(2, "no comands specified");
			exit_error(2,"no time support in LIDS 2.2\n");
			break;
		case 'i':
			if (command == LIDS_NONE)
				exit_error(2, "no comands specified");
			inherit = atoi(optarg);
			break;
		case 'd':
			exit_error(2, "domain not supported now");
			break;
		case 'j':
			if (command == LIDS_NONE)
				exit_error(2, "no comands specified");

			if (type != LIDS_NONE)
				exit_error(2, "multiple types specified");
			type = -1;

			for (i = 0; i < LIDS_TARGET_NUM; i++) {
				if (!strcmp(optarg, lids_type_desc[i])) {
					type = lids_type_val[i];
				}
			}
			if (type == -1)
				exit_error(2,
					   "cap type must be READONLY, WRITE, APPEND, DENY , ENABLE, DISABLE or GRANT");
			break;
		case 'P':
			if (command != LIDS_NONE)
				exit_error(2, "multiple commands specified");
			command = LIDS_MK_PASSWD;
			break;
		case 'C':
			if (command != LIDS_NONE)
				exit_error(2, "multiple commands specified");
			command = LIDS_CHECK;
			break;
		case 'V':
			verbose = 1;
			break;
		case 'v':
			exit_version();
			break;
		case 'h':
			exit_help(0);
		case 'H':
			exit_help(1);
		default:
			break;
		}
	}
	if (optind < argc)
		exit_error(2, "unknown arguments found on commandline");
	if ((command == LIDS_NONE) || (argc < 2))
		exit_normal();
	if ((command == LIDS_ADD) && (type == LIDS_NONE))
		exit_normal();
	if ((subject_file[0] != 0) && (object_file[0] == 0)
	    && (command == LIDS_ADD))
		exit_error(2, "defined subject must be related to an object");

	switch (command) {
	case LIDS_ADD:
		verbose_printf("ADD\n");
		lids_add_file(subject_file, object_file, type, inherit,
			      timescale);
		fprintf(stderr,"Added 1 ACL\n");
		break;
	case LIDS_DELETE:
		verbose_printf("DELETE\n");
		lids_del_file(subject_file, object_file);
		break;
	case LIDS_LIST:
		verbose_printf("LIST\n");
		lids_list_file(type);
		break;
 	case LIDS_SCRIPT:
 		verbose_printf("SCRIPT\n");
 		lids_write_script();
 		break;
	case LIDS_DELETE_ALL:
		verbose_printf("DELETE_ALL\n");
		lids_del_all();
		break;
	case LIDS_MK_PASSWD:
		verbose_printf("MAKE PASSWD\n");
		lids_make_passwd();
		break;
	case LIDS_UPDATE:
		verbose_printf("UPDATE\n");
		lids_update();
		break;
	case LIDS_CHECK:
		verbose_printf("INTERIGITY CHECK\n");
		verbose = 1;
		lids_check();
		break;
	}
	exit(0);
}
